import json
import pathlib
import shutil
from itertools import product
from pathlib import Path
from pickle import dumps, loads

import pytest
from citeable import Software
from scitrack import get_text_hexdigest

try:
    import cogent3 as c3
    from cogent3.util.union_dict import UnionDict
except ImportError:
    c3 = None
    UnionDict = None

from scinexus import open_data_store
from scinexus.composable import NotCompleted, NotCompletedType
from scinexus.data_store import (
    APPEND,
    CITATIONS_FILE,
    MD5_TABLE,
    NOT_COMPLETED_TABLE,
    OVERWRITE,
    READONLY,
    DataStoreDirectory,
    ReadOnlyDataStoreZipped,
    get_data_source,
    get_id_from_source,
    get_summary_display,
    get_unique_id,
    load_record_from_json,
    make_record_for_json,
    set_id_from_source,
    set_summary_display,
    summary_not_completeds,
)

# over-ride cogent3 setting
set_summary_display(None)


@pytest.fixture
def tmp_dir(tmp_path_factory):
    return Path(tmp_path_factory.mktemp("datastore"))


@pytest.fixture
def fasta_dir(DATA_DIR, tmp_dir):
    tmp_dir = Path(tmp_dir)
    filenames = DATA_DIR.glob("*.fasta")
    fasta_dir = tmp_dir / "fasta"
    fasta_dir.mkdir(parents=True, exist_ok=True)
    for fn in filenames:
        dest = fasta_dir / fn.name
        dest.write_text(fn.read_text())
    return fasta_dir


@pytest.fixture
def write_dir(tmp_dir):
    tmp_dir = Path(tmp_dir)
    write_dir = tmp_dir / "write"
    write_dir.mkdir(parents=True, exist_ok=True)
    yield write_dir
    shutil.rmtree(write_dir, ignore_errors=True)


@pytest.fixture
def w_dstore(write_dir):
    return DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)


@pytest.fixture
def ro_dstore(fasta_dir):
    return DataStoreDirectory(fasta_dir, suffix="fasta", mode=READONLY)


@pytest.fixture
def completed_objects(ro_dstore):
    return {f"{Path(m.unique_id).stem}": m.read() for m in ro_dstore}


@pytest.fixture
def nc_objects():
    return {
        f"id_{i}": NotCompleted(
            NotCompletedType.ERROR, "location", "message", source=f"id_{i}"
        )
        for i in range(3)
    }


@pytest.fixture(scope="session")
def log_data(DATA_DIR):
    path = DATA_DIR / "scitrack.log"
    return path.read_text()


@pytest.fixture
def full_dstore(write_dir, nc_objects, completed_objects, log_data):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    for id_, data in nc_objects.items():
        dstore.write_not_completed(unique_id=id_, data=data.to_json())

    for id_, data in completed_objects.items():
        dstore.write(unique_id=id_, data=data)

    dstore.write_log(unique_id="scitrack.log", data=log_data)
    return dstore


@pytest.fixture
def nc_dir(tmp_dir):
    nc_dir = tmp_dir / "nc_test"
    nc_dir.mkdir(parents=True, exist_ok=True)
    yield nc_dir
    shutil.rmtree(nc_dir, ignore_errors=True)


@pytest.fixture
def nc_dstore(DATA_DIR, nc_dir):
    dstore = DataStoreDirectory(nc_dir, suffix="fasta", mode=OVERWRITE)
    log_filename = "scitrack.log"
    dstore.write_log(unique_id=log_filename, data=(DATA_DIR / log_filename).read_text())
    nc = [
        NotCompleted(
            NotCompletedType.FAIL,
            f"dummy{i}",
            f"dummy_message{i}",
            source=f"dummy_source{i}",
        )
        for i in range(3)
    ]
    for i, item in enumerate(nc):
        dstore.write_not_completed(unique_id=f"nc{i + 1}", data=item.to_json())
    assert len(dstore.not_completed) == 3
    filenames = DATA_DIR.glob("*.fasta")
    for fn in filenames:
        identifier = fn.name
        dstore.write(unique_id=identifier, data=fn.read_text())
    return dstore


@pytest.fixture
def sample_citations():
    cite1 = Software(
        author=["Doe, J"],
        title="Tool One",
        year=2024,
        url="https://example.com/one",
        version="1.0",
        license="MIT",
        doi="10.0/one",
        publisher="test",
    )
    cite2 = Software(
        author=["Smith, A"],
        title="Tool Two",
        year=2024,
        url="https://example.com/two",
        version="2.0",
        license="MIT",
        doi="10.0/two",
        publisher="test",
    )
    return (cite1, cite2)


def _get_member_data(members):
    return {m.unique_id: m.read() for m in members}


@pytest.fixture
def zipped_basic(fasta_dir):
    path = shutil.make_archive(
        base_name=str(fasta_dir.parent / fasta_dir.name),
        format="zip",
        base_dir=fasta_dir.name,
        root_dir=fasta_dir.parent,
    )
    return pathlib.Path(path)


@pytest.fixture
def zipped_full(full_dstore):
    source = pathlib.Path(full_dstore.source)
    path = shutil.make_archive(
        base_name=str(source.parent / source.name),
        format="zip",
        base_dir=source.name,
        root_dir=source.parent,
    )
    return ReadOnlyDataStoreZipped(pathlib.Path(path), suffix="fasta")


@pytest.fixture
def zipped_hidden(fasta_dir):
    # create a hidden file
    hidden = fasta_dir / ".hidden.fasta"
    hidden.write_text(">s1\nACGT\n")
    path = shutil.make_archive(
        base_name=str(fasta_dir.parent / (fasta_dir.name + "_hidden")),
        format="zip",
        base_dir=fasta_dir.name,
        root_dir=fasta_dir.parent,
    )
    hidden.unlink()
    return pathlib.Path(path)


def test_data_member_eq(ro_dstore, fasta_dir):
    ro_dstore2 = DataStoreDirectory(fasta_dir, mode="r", suffix="fasta")
    name = "brca1.fasta"
    mem1 = next(m for m in ro_dstore.completed if m.unique_id == name)
    mem2 = next(m for m in ro_dstore2.completed if m.unique_id == name)
    assert mem1 != mem2


def test_fail_try_append(full_dstore, completed_objects):
    full_dstore._mode = APPEND
    id_, data = next(iter(completed_objects.items()))
    with pytest.raises(IOError):
        full_dstore.write(unique_id=id_, data=data)


def test_contains(ro_dstore):
    """correctly identify when a data store contains a member"""
    assert "brca1.fasta" in ro_dstore
    assert "brca1" in ro_dstore


def test_len(ro_dstore):
    """DataStore returns correct len"""
    expect = len(list(ro_dstore.source.glob("*.fasta")))
    assert expect == len(ro_dstore) == len(ro_dstore.members)


def test_getitem(ro_dstore):
    with pytest.raises(IndexError):
        _ = ro_dstore[len(ro_dstore)]

    last = ro_dstore[-1]
    first = ro_dstore[0]
    assert last.unique_id != first.unique_id


def test_iterall(ro_dstore):
    expect = {fn.name for fn in ro_dstore.source.glob("*.fasta")}
    got = {m.unique_id for m in ro_dstore}
    assert expect == got


def test_read(ro_dstore):
    """correctly read content"""
    expect = (ro_dstore.source / "brca1.fasta").read_text()
    got = ro_dstore.read("brca1.fasta")
    assert got == expect


def test_pickleable_roundtrip(ro_dstore):
    """pickling of data stores should be reversible"""
    re_dstore = loads(dumps(ro_dstore))
    assert str(ro_dstore) == str(re_dstore)
    assert ro_dstore[0].read() == re_dstore[0].read()


def test_pickleable_member_roundtrip(ro_dstore):
    """pickling of data store members should be reversible"""
    re_member = loads(dumps(ro_dstore[0]))
    data = re_member.read()
    assert len(data) > 0


def test_empty_directory(fasta_dir):
    dstore = DataStoreDirectory(fasta_dir, suffix=".txt")
    assert len(dstore) == 0


def test_no_logs(ro_dstore):
    assert len(ro_dstore.logs) == 0


def test_no_not_completed(ro_dstore):
    assert len(ro_dstore.not_completed) == 0


def test_logs(nc_dstore):
    assert len(nc_dstore.logs) == 1
    log = nc_dstore.logs[0].read()
    assert isinstance(log, str)


def test_not_completed(nc_dstore):
    assert len(nc_dstore.not_completed) == 3
    nc = nc_dstore.not_completed[0].read()
    assert isinstance(nc, str)


def test_drop_not_completed(nc_dstore):
    num_completed = len(nc_dstore.completed)
    num_not_completed = len(nc_dstore.not_completed)
    num_md5 = len(list((nc_dstore.source / MD5_TABLE).glob("*.txt")))
    assert num_not_completed == 3
    assert num_completed == 6
    assert len(nc_dstore) == 9
    assert num_md5 == num_completed + num_not_completed
    nc_dstore.drop_not_completed()
    assert len(nc_dstore.not_completed) == 0
    num_md5 = len(list((nc_dstore.source / MD5_TABLE).glob("*.txt")))
    assert num_md5 == num_completed


def test_write_read_only_datastore(ro_dstore):
    with pytest.raises(IOError):
        ro_dstore.write(unique_id="brca1.fasta", data="test data")


def test_write(fasta_dir, w_dstore):
    """correctly write content"""
    expect = Path(fasta_dir / "brca1.fasta").read_text()
    identifier = "brca1.fasta"
    w_dstore.write(unique_id=identifier, data=expect)
    got = w_dstore.read(identifier)
    assert got == expect


def test_multi_write(fasta_dir, w_dstore):
    """correctly write multiple files to data store"""
    expect_a = Path(fasta_dir / "brca1.fasta").read_text()
    expect_b = Path(fasta_dir / "primates_brca1.fasta").read_text()
    identifier_a = "brca2.fasta"
    identifier_b = "primates_brca2.fasta"
    w_dstore.write(unique_id=identifier_a, data=expect_a)
    w_dstore.write(unique_id=identifier_b, data=expect_b)
    got_a = w_dstore.read(identifier_a)
    got_b = w_dstore.read(identifier_b)
    assert got_a == expect_a
    assert got_b == expect_b


def test_append(w_dstore):
    """correctly write content"""
    identifier = "test1.fasta"
    data = "test data"
    w_dstore.write(unique_id=identifier, data=data)
    got = w_dstore.read(identifier)
    assert got == data


def test_no_not_completed_subdir(nc_dstore):
    expect = f"{len(nc_dstore.completed) + len(nc_dstore.not_completed)}x member"
    assert str(nc_dstore).startswith(expect)
    nc_dstore.drop_not_completed()
    assert not Path(nc_dstore.source / NOT_COMPLETED_TABLE).exists()
    expect = f"{len(nc_dstore.completed)}x member"
    assert str(nc_dstore).startswith(expect)
    expect = f"{len(nc_dstore)}x member"
    assert str(nc_dstore).startswith(expect)
    assert len(nc_dstore) == len(nc_dstore.completed)
    not_dir = nc_dstore.source / NOT_COMPLETED_TABLE
    not_dir.mkdir(exist_ok=True)


def test_limit_datastore(nc_dstore):
    assert len(nc_dstore) == len(nc_dstore.completed) + len(nc_dstore.not_completed)
    nc_dstore._limit = len(nc_dstore.completed) // 2
    nc_dstore._completed = []
    nc_dstore._not_completed = []
    assert len(nc_dstore.completed) == len(nc_dstore.not_completed) == nc_dstore.limit
    assert len(nc_dstore) == len(nc_dstore.completed) + len(nc_dstore.not_completed)
    nc_dstore.drop_not_completed()
    assert len(nc_dstore) == len(nc_dstore.completed)
    assert len(nc_dstore.not_completed) == 0
    nc_dstore._limit = len(nc_dstore.completed) // 2
    nc_dstore._completed = []
    nc_dstore._not_completed = []
    assert len(nc_dstore) == len(nc_dstore.completed) == nc_dstore.limit
    assert len(nc_dstore.not_completed) == 0


def test_md5_sum(nc_dstore):
    for m in nc_dstore.members:
        data = m.read()
        md5 = nc_dstore.md5(m.unique_id)
        assert md5 == get_text_hexdigest(data)


def test_md5_none(fasta_dir):
    dstore = DataStoreDirectory(fasta_dir, suffix="fasta")
    for m in dstore.members:
        assert m.md5 is None


def test_md5_missing(nc_dstore):
    assert nc_dstore.md5("unknown") is None


def test_write_if_member_exists(full_dstore, write_dir):
    """correctly write content"""
    expect = Path(write_dir / "brca1.fasta").read_text()
    identifier = "brca1.fasta"
    len_dstore = len(full_dstore)
    full_dstore.write(unique_id=identifier, data=expect)
    assert len_dstore == len(full_dstore)
    got = full_dstore.read(identifier)
    assert got == expect
    full_dstore._mode = OVERWRITE
    full_dstore.write(unique_id=identifier, data=expect)
    assert len_dstore == len(full_dstore)
    got = full_dstore.read(identifier)
    assert got == expect


def test_write_success_replaces_not_completed(full_dstore):
    """correctly write content"""
    nc = full_dstore.not_completed[0].unique_id
    data = full_dstore.completed[0].read()
    new_id = Path(nc.replace(".json", f".{full_dstore.suffix}")).name
    num = len(full_dstore)
    full_dstore.write(unique_id=new_id, data=data)
    assert len(full_dstore) == num


@pytest.mark.parametrize("klass", [str, Path])
def test_get_data_source_attr(klass):
    """handles case where input has source attribute string object or pathlib object"""

    class dummy:
        source = None

    obj = dummy()
    value = klass("some/path.txt")
    obj.source = value
    got = get_data_source(obj)
    assert got == "path.txt"


@pytest.mark.parametrize(
    "name",
    ["path/name.txt", "path/name.gz", "path/name.fasta.gz", "name.fasta.gz"],
)
def test_get_unique_id(name):
    got = get_unique_id(name)
    assert got == "name"


def test_get_unique_id_none():
    got = get_unique_id(None)
    assert got is None


def test_set_id_from_source_returns_default_initially(
    reset_id_from_source: None,
) -> None:
    """Default extractor is `get_unique_id` when nothing is registered."""
    assert get_id_from_source() is get_unique_id


def test_set_id_from_source_registers_and_clears(
    reset_id_from_source: None,
) -> None:
    """A registered function replaces the default; None restores it."""

    def my_extractor(obj: object) -> str | None:
        return f"custom-{obj}"

    set_id_from_source(my_extractor)
    assert get_id_from_source() is my_extractor
    assert get_id_from_source()("foo") == "custom-foo"

    set_id_from_source(None)
    assert get_id_from_source() is get_unique_id


@pytest.mark.parametrize("data", [{}, set(), {"info": {}}])
def test_get_data_source_none(data):
    assert get_data_source(data) is None


def test_load_record_from_json():
    """handle different types of input"""
    orig = {"data": "blah", "identifier": "some.json", "completed": True}
    data = orig.copy()
    data2 = data.copy()
    data2["data"] = json.dumps(data)
    for d in (data, json.dumps(data), data2):
        expected = "blah" if d != data2 else json.loads(data2["data"])
        id_, data_, compl = load_record_from_json(d)
        assert id_ == "some.json"
        assert data_ == expected
        assert compl is True


def test_zipped_ro_fail(zipped_basic):
    with pytest.raises(ValueError):
        ReadOnlyDataStoreZipped(zipped_basic, suffix="fasta", mode="w")


def test_zipped_ro_ioerror():
    with pytest.raises(IOError):
        ReadOnlyDataStoreZipped("blah-1234.zip", suffix="fasta")


def test_zipped_ro_basic(zipped_basic, ro_dstore):
    dstore = ReadOnlyDataStoreZipped(zipped_basic, suffix="fasta")
    assert len(dstore.completed) == len(ro_dstore.completed)
    assert len(dstore) == len(ro_dstore)
    expect = _get_member_data(ro_dstore.completed)
    got = _get_member_data(dstore.completed)
    assert expect == got
    expect = _get_member_data(ro_dstore.not_completed)
    got = _get_member_data(dstore.not_completed)
    assert expect == got


def test_zipped_ro_basic_hidden(zipped_hidden, zipped_basic):
    orig = ReadOnlyDataStoreZipped(zipped_basic, suffix="fasta")
    dstore = ReadOnlyDataStoreZipped(zipped_hidden, suffix="fasta")
    assert len(dstore) == len(orig)
    assert all(not m.unique_id.startswith(".") for m in dstore)


def test_zipped_ro_full(zipped_full, full_dstore):
    got_ids = {m.unique_id for m in zipped_full.completed}
    expect_ids = {m.unique_id for m in full_dstore.completed}
    assert got_ids == expect_ids

    got_ids = {m.unique_id for m in zipped_full.not_completed}
    expect_ids = {m.unique_id for m in full_dstore.not_completed}
    assert got_ids == expect_ids
    assert len(zipped_full) == len(full_dstore)

    expect = _get_member_data(full_dstore.completed)
    got = _get_member_data(zipped_full.completed)
    assert expect == got
    expect = _get_member_data(full_dstore.not_completed)
    got = _get_member_data(zipped_full.not_completed)
    assert expect == got


def test_zipped_logs(zipped_full, full_dstore):
    assert len(zipped_full.logs) == len(full_dstore.logs)
    expect = _get_member_data(full_dstore.logs)
    got = _get_member_data(zipped_full.logs)
    assert expect == got


def test_zipped_md5(zipped_full, full_dstore):
    expect = {m.unique_id: full_dstore.md5(m.unique_id) for m in full_dstore.completed}
    got = {m.unique_id: zipped_full.md5(m.unique_id) for m in zipped_full.completed}
    assert got == expect


def test_write_citations_directory(write_dir, sample_citations):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    dstore.write_citations(data=sample_citations)
    path = write_dir / CITATIONS_FILE
    assert path.exists()
    loaded = dstore._load_citations()
    assert len(loaded) == 2
    assert loaded[0].title == "Tool One"
    assert loaded[1].title == "Tool Two"


def test_write_citations_empty_directory(write_dir):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    dstore.write_citations(data=())
    path = write_dir / CITATIONS_FILE
    assert not path.exists()


def test_write_bib_directory(write_dir, sample_citations):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    dstore.write_citations(data=sample_citations)
    bib_path = write_dir / "refs.bib"
    dstore.write_bib(bib_path)
    assert bib_path.exists()
    content = bib_path.read_text()
    assert "Tool One" in content
    assert "Tool Two" in content


def test_write_bib_no_citations(write_dir):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    bib_path = write_dir / "refs.bib"
    with pytest.warns(UserWarning, match="No citations stored"):
        dstore.write_bib(bib_path)
    assert not bib_path.exists()


def test_load_citations_no_file(write_dir):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    assert dstore._load_citations() == []


def test_load_citations_zipped(write_dir, sample_citations):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    dstore.write_citations(data=sample_citations)
    source = pathlib.Path(dstore.source)
    path = shutil.make_archive(
        base_name=str(source.parent / source.name),
        format="zip",
        base_dir=source.name,
        root_dir=source.parent,
    )
    zipped = ReadOnlyDataStoreZipped(pathlib.Path(path), suffix="fasta")
    loaded = zipped._load_citations()
    assert len(loaded) == 2
    assert loaded[0].title == "Tool One"


def test_citations_file_not_in_completed(write_dir, sample_citations):
    """The bibliography.citations file must not appear in the completed members list."""
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    dstore.write(unique_id="sample.fasta", data=">s1\nACGT\n")
    dstore.write_citations(data=sample_citations)
    assert (write_dir / CITATIONS_FILE).exists()
    dstore._completed = []
    member_ids = {m.unique_id for m in dstore.completed}
    assert CITATIONS_FILE not in member_ids
    assert "sample.fasta" in member_ids


@pytest.fixture
def _restore_display():
    """Ensure the global display function is reset after each test."""
    yield
    set_summary_display(None)


def test_summary_display_default_is_none(_restore_display):
    assert get_summary_display() is None


def test_summary_display_set_and_get(_restore_display):
    def my_display(data, *, name=""):
        return data

    set_summary_display(my_display)
    assert get_summary_display() is my_display


def test_summary_display_set_none_clears(_restore_display):
    set_summary_display(lambda data, **kw: data)
    set_summary_display(None)
    assert get_summary_display() is None


def test_describe_without_display(ro_dstore, _restore_display):
    result = ro_dstore.describe
    assert isinstance(result, dict)
    assert "completed" in result


def test_describe_with_display(ro_dstore, _restore_display):
    captured = {}

    def display(data, *, name=""):
        captured["data"] = data
        captured["name"] = name
        return f"DISPLAY:{name}"

    set_summary_display(display)
    result = ro_dstore.describe
    assert result == "DISPLAY:describe"
    assert isinstance(captured["data"], dict)
    assert "completed" in captured["data"]
    assert captured["name"] == "describe"


def test_summary_logs_with_display(full_dstore, _restore_display):
    captured = {}

    def display(data, *, name=""):
        captured["data"] = data
        captured["name"] = name
        return "transformed"

    set_summary_display(display)
    result = full_dstore.summary_logs
    assert result == "transformed"
    assert captured["name"] == "summary_logs"
    assert isinstance(captured["data"], list)


def test_validate_with_display(ro_dstore, _restore_display):
    captured = {}

    def display(data, *, name=""):
        captured["name"] = name
        return "validated"

    set_summary_display(display)
    result = ro_dstore.validate()
    assert result == "validated"
    assert captured["name"] == "validate"


def test_protected_methods_bypass_display(ro_dstore, _restore_display):
    set_summary_display(lambda data, **kw: "SHOULD_NOT_SEE")
    assert isinstance(ro_dstore._describe(), dict)
    assert isinstance(ro_dstore._summary_logs(), list)
    assert isinstance(ro_dstore._summary_not_completed(), list)
    assert isinstance(ro_dstore._validate(), dict)


def test_summary_citations_with_display(write_dir, sample_citations, _restore_display):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    dstore.write_citations(data=sample_citations)
    captured = {}

    def display(data, *, name=""):
        captured["name"] = name
        captured["data"] = data
        return "citations_display"

    set_summary_display(display)
    result = dstore.summary_citations
    assert result == "citations_display"
    assert captured["name"] == "summary_citations"
    assert isinstance(captured["data"], list)


def test_validate_incorrect_md5(write_dir):
    dstore = DataStoreDirectory(write_dir, suffix="txt", mode=OVERWRITE)
    dstore.write(unique_id="item.txt", data="original")
    # corrupt the md5
    md5_path = write_dir / MD5_TABLE / "item.txt"
    md5_path.write_text("wrong_md5_value")
    result = dstore._validate()
    assert result["md5_incorrect"] == 1


def test_readonly_nonexistent_dir(tmp_path):
    with pytest.raises(OSError, match="does not exist"):
        DataStoreDirectory(tmp_path / "nonexistent", suffix="txt", mode=READONLY)


def test_not_completed_with_limit(write_dir):
    dstore = DataStoreDirectory(write_dir, suffix="txt", mode=OVERWRITE)
    for i in range(5):
        nc = NotCompleted(NotCompletedType.ERROR, "test", f"msg {i}", source=f"src_{i}")
        dstore.write_not_completed(unique_id=f"nc_{i}.json", data=nc.to_json())
    limited = DataStoreDirectory(write_dir, suffix="txt", mode=READONLY, limit=2)
    assert len(limited.not_completed) == 2


def test_summary_not_completeds(write_dir):
    dstore = DataStoreDirectory(write_dir, suffix="txt", mode=OVERWRITE)
    for i in range(3):
        nc = NotCompleted(
            NotCompletedType.ERROR, "myapp", f"error msg {i}", source=f"s{i}"
        )
        dstore.write_not_completed(unique_id=f"nc_{i}.json", data=nc.to_json())
    rows = summary_not_completeds(dstore.not_completed)
    assert len(rows) >= 1
    assert rows[0]["origin"] == "myapp"
    assert rows[0]["num"] == 3


def test_make_record_for_json():
    result = make_record_for_json("id1", {"key": "value"}, True)
    assert result["identifier"] == "id1"
    assert result["completed"] is True
    assert isinstance(result["data"], str)
    parsed = json.loads(result["data"])
    assert parsed == {"key": "value"}


def test_make_record_for_json_with_rich_dict():
    class FakeObj:
        def to_rich_dict(self):
            return {"type": "fake", "data": 42}

    result = make_record_for_json("id2", FakeObj(), True)
    parsed = json.loads(result["data"])
    assert parsed == {"type": "fake", "data": 42}


def test_zipped_readonly_write_methods(tmp_path):
    src = tmp_path / "src"
    src.mkdir()
    (src / "a.txt").write_text("data")
    zpath = shutil.make_archive(str(src), "zip", root_dir=src.parent, base_dir=src.name)
    zstore = ReadOnlyDataStoreZipped(zpath, suffix="txt")

    with pytest.raises(TypeError):
        zstore.write(unique_id="x", data="d")
    with pytest.raises(TypeError):
        zstore.write_not_completed(unique_id="x", data="d")
    with pytest.raises(TypeError):
        zstore.write_log(unique_id="x", data="d")
    with pytest.raises(TypeError):
        zstore.write_citations(data=())
    with pytest.raises(TypeError):
        zstore.drop_not_completed()


def test_zipped_md5_returns_none(tmp_path):
    src = tmp_path / "src"
    src.mkdir()
    (src / "a.txt").write_text("data")
    zpath = shutil.make_archive(str(src), "zip", root_dir=src.parent, base_dir=src.name)
    zstore = ReadOnlyDataStoreZipped(zpath, suffix="txt")
    assert zstore.md5("a.txt") is None


def test_zipped_load_citations_missing(tmp_path):
    src = tmp_path / "src"
    src.mkdir()
    (src / "a.txt").write_text("data")
    zpath = shutil.make_archive(str(src), "zip", root_dir=src.parent, base_dir=src.name)
    zstore = ReadOnlyDataStoreZipped(zpath, suffix="txt")
    assert zstore._load_citations() == []


def test_zipped_completed_with_limit(tmp_path):
    src = tmp_path / "src"
    src.mkdir()
    for i in range(5):
        (src / f"f_{i}.txt").write_text(f"data {i}")
    zpath = shutil.make_archive(str(src), "zip", root_dir=src.parent, base_dir=src.name)
    zstore = ReadOnlyDataStoreZipped(zpath, suffix="txt", limit=2)
    assert len(zstore.completed) == 2


def test_summary_logs_continuation_line(write_dir):
    from scitrack import CachingLogger

    dstore = DataStoreDirectory(write_dir, suffix="txt", mode=OVERWRITE)
    logger = CachingLogger(create_dir=True)
    log_path = write_dir / "test.log"
    logger.log_file_path = str(log_path)
    logger.log_message("a long message\nthat continues", label="multi")
    logger.shutdown()
    dstore.write_log(unique_id="test.log", data=log_path.read_text())
    rows = dstore._summary_logs()
    assert len(rows) == 1


def test_data_member_str(ro_dstore):
    member = ro_dstore[0]
    assert str(member) == member.unique_id


def test_data_member_repr(ro_dstore):
    member = ro_dstore[0]
    r = repr(member)
    assert "DataMember" in r
    assert member.unique_id in r


def test_datastore_repr(ro_dstore):
    r = repr(ro_dstore)
    assert "DataStoreDirectory" in r
    assert "source=" in r


def test_contains_non_string(ro_dstore):
    assert 42 not in ro_dstore


def test_write_not_completed_readonly(ro_dstore):
    with pytest.raises(OSError, match="readonly"):
        ro_dstore.write_not_completed(unique_id="x", data="d")


def test_write_log_readonly(ro_dstore):
    with pytest.raises(OSError, match="readonly"):
        ro_dstore.write_log(unique_id="x", data="d")


def test_summary_logs_malformed_continuation():
    from unittest.mock import MagicMock

    log_text = "2024-01-01\t00:00:00\n\tcontinuation without key"
    member = MagicMock()
    member.read.return_value = log_text
    member.unique_id = "bad.log"

    class FakeDS(DataStoreDirectory):
        @property
        def logs(self):
            return [member]

    fake = FakeDS.__new__(FakeDS)
    fake._completed = []
    fake._not_completed = []
    fake._init_vals = {}
    with pytest.raises(ValueError, match="malformed log data"):
        fake._summary_logs()


def test_tidy_and_check_suffix_empty():
    from scinexus.data_store import _tidy_and_check_suffix

    with pytest.raises(ValueError, match="suffix is required"):
        _tidy_and_check_suffix(None)

    with pytest.raises(ValueError, match="suffix is required"):
        _tidy_and_check_suffix("*")

    with pytest.raises(ValueError, match="suffix is required"):
        _tidy_and_check_suffix(".**")


def test_summary_not_completeds_with_bytes():
    from unittest.mock import MagicMock

    members = []
    for i in range(3):
        m = MagicMock()
        m.read.return_value = b"binary data"
        m.unique_id = f"item_{i}"
        members.append(m)
    result = summary_not_completeds(members)
    assert result == []


def test_summary_not_completeds_with_deserialise(write_dir):
    dstore = DataStoreDirectory(write_dir, suffix="txt", mode=OVERWRITE)
    from scinexus.composable import NotCompleted, NotCompletedType

    for i in range(3):
        nc = NotCompleted(
            NotCompletedType.ERROR, "myapp", f"error msg {i}", source=f"s{i}"
        )
        dstore.write_not_completed(unique_id=f"deser_{i}.json", data=nc.to_json())
    rows = summary_not_completeds(dstore.not_completed, deserialise=lambda x: x)
    assert len(rows) >= 1


def test_summary_not_completeds_long_sources(write_dir):
    dstore = DataStoreDirectory(write_dir, suffix="txt", mode=OVERWRITE)
    from scinexus.composable import NotCompleted, NotCompletedType

    for i in range(10):
        long_source = f"very_long_source_name_for_item_{i}_padding"
        nc = NotCompleted(
            NotCompletedType.ERROR,
            "myapp",
            f"error msg {i}",
            source=long_source,
        )
        dstore.write_not_completed(unique_id=f"long_{i}.json", data=nc.to_json())
    rows = summary_not_completeds(dstore.not_completed)
    assert len(rows) >= 1
    for row in rows:
        if len(row["source"]) > 45:
            assert row["source"].endswith("...")


def test_get_data_source_data_member(ro_dstore):
    member = ro_dstore[0]
    result = get_data_source(member)
    assert result == member.unique_id


def test_zipped_not_completed_with_limit(tmp_path):
    from scinexus.composable import NotCompleted, NotCompletedType

    src = tmp_path / "src"
    src.mkdir()
    nc_dir = src / "not_completed"
    nc_dir.mkdir()
    (src / "a.txt").write_text("data")
    for i in range(5):
        nc = NotCompleted(NotCompletedType.ERROR, "test", f"msg {i}", source=f"s{i}")
        (nc_dir / f"nc_{i}.json").write_text(nc.to_json())
    zpath = shutil.make_archive(str(src), "zip", root_dir=src.parent, base_dir=src.name)
    zstore = ReadOnlyDataStoreZipped(zpath, suffix="txt", limit=2)
    assert len(zstore.not_completed) == 2


def test_zipped_mode_property(zipped_basic):
    zstore = ReadOnlyDataStoreZipped(zipped_basic, suffix="fasta")
    assert zstore.mode is READONLY


def _make_minimal_ds(mode=OVERWRITE):
    from scinexus.data_store import DataStoreABC

    class MinimalDS(DataStoreABC):
        @property
        def source(self):
            return "test"

        @property
        def mode(self):
            return mode

        @property
        def limit(self):
            return None

        def read(self, unique_id):
            return ""

        def write(self, *, unique_id, data):
            super().write(unique_id=unique_id, data=data)

        def write_not_completed(self, *, unique_id, data):
            super().write_not_completed(unique_id=unique_id, data=data)

        def write_log(self, *, unique_id, data):
            super().write_log(unique_id=unique_id, data=data)

        @property
        def logs(self):
            return []

        @property
        def completed(self):
            return []

        @property
        def not_completed(self):
            return []

        def drop_not_completed(self, *, unique_id=None):
            pass

        def md5(self, unique_id):
            return None

    return MinimalDS()


def test_base_write_citations_warns():
    ds = _make_minimal_ds()
    with pytest.warns(UserWarning, match="does not support saving citations"):
        ds.write_citations(data=(object(),))


def test_base_summary_citations_warns():
    ds = _make_minimal_ds()
    with pytest.warns(UserWarning, match="does not support saving citations"):
        result = ds._summary_citations()
    assert result == []


def test_base_load_citations_returns_empty():
    ds = _make_minimal_ds()
    assert ds._load_citations() == []


def test_base_write_not_completed_readonly():
    ds = _make_minimal_ds(mode=READONLY)
    with pytest.raises(OSError, match="readonly"):
        ds.write_not_completed(unique_id="x", data="d")


def test_base_write_log_readonly():
    ds = _make_minimal_ds(mode=READONLY)
    with pytest.raises(OSError, match="readonly"):
        ds.write_log(unique_id="x", data="d")


def test_base_write_citations_empty_data():
    ds = _make_minimal_ds()
    ds.write_citations(data=())


@pytest.mark.mpi
def test_source_check_create_not_master(tmp_path):
    from unittest.mock import patch

    from scinexus import data_store as ds_mod

    target = tmp_path / "should_not_exist"
    with patch.object(ds_mod, "is_master_process", return_value=False):
        dstore = DataStoreDirectory(target, suffix="txt", mode=OVERWRITE)
    assert not target.exists()
    assert dstore.source == target


def test_write_read_not_completed(nc_dstore):
    nc_dstore.drop_not_completed()
    assert len(nc_dstore.not_completed) == 0
    nc = NotCompleted("ERROR", "test", "for tracing", source="blah")
    writer = c3.get_app("write_seqs", data_store=nc_dstore)
    writer.main(nc, identifier="blah")
    assert len(nc_dstore.not_completed) == 1
    got = nc_dstore.not_completed[0].read()
    assert nc.to_json() == got


def test_summary_logs_missing_field(nc_dstore):
    log_path = Path(nc_dstore.source) / nc_dstore.logs[0].unique_id
    data = [
        l for l in log_path.read_text().splitlines() if "composable function" not in l
    ]
    log_path.write_text("\n".join(data))
    # doesn't fail because of a missing field in the log data
    assert isinstance(nc_dstore.summary_logs, list)


@pytest.fixture
def app_dstore_in(tmp_path):
    pytest.importorskip("cogent3")
    in_path = tmp_path / "in_data"
    in_path.mkdir(parents=True)
    fasta_content = ">seq\nACGT"
    with open(in_path / "one.fa", "w") as file:
        file.write(fasta_content)

    dstore_in = open_data_store(in_path, suffix=".fa", mode="r")
    dstore_out = open_data_store(tmp_path / "data_out", suffix="fa", mode="w")
    loader = c3.get_app("load_unaligned")
    writer = c3.get_app("write_seqs", dstore_out)

    pipe = loader + writer
    return pipe, dstore_in


def test_write_multiple_times_apply_to(app_dstore_in):
    app, dstore_in = app_dstore_in
    app.apply_to(dstore_in)
    orig_length = len(app.data_store)
    app.apply_to(dstore_in)
    assert len(app.data_store) == orig_length


def test_directory_data_store_write_compressed(tmp_path):
    out = open_data_store(base_path=tmp_path / "demo", suffix="fa.gz", mode="w")
    writer = c3.get_app("write_seqs", data_store=out)
    seqs = c3.make_aligned_seqs(
        {"s1": "CG--T", "s2": "CGTTT"},
        moltype="dna",
        info={"source": "test"},
    )
    got = writer(seqs)  # pylint: disable=not-callable
    assert got, got


def test_apply_to_not_completed(nc_dstore, tmp_path):
    loader = c3.get_app("load_unaligned")
    num_seqs = c3.get_app("take_n_seqs", number=3, fixed_choice=False)
    out_dstore = open_data_store(tmp_path / "output", suffix="fa", mode="w")
    writer = c3.get_app("write_seqs", data_store=out_dstore, format_name="fasta")
    app = loader + num_seqs + writer
    fini = app.apply_to(nc_dstore)
    assert 0 < len(fini.completed) <= len(nc_dstore.completed)


def test_summary_citations_directory(write_dir, sample_citations):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    dstore.write_citations(data=sample_citations)
    cited = dstore.summary_citations
    assert isinstance(cited, list)
    assert len(cited) == 2
    assert "app" in cited[0]
    assert "citation" in cited[0]


def test_write_bib_tilde_path(write_dir, sample_citations, HOME_TMP_DIR):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    dstore.write_citations(data=sample_citations)
    bib_path = f"~/{HOME_TMP_DIR.name}/refs.bib"
    dstore.write_bib(bib_path)
    expected = pathlib.Path(bib_path).expanduser()
    assert expected.exists()
    content = expected.read_text()
    assert "Tool One" in content
    assert "Tool Two" in content


def test_summary_citations_zipped(write_dir, sample_citations):
    dstore = DataStoreDirectory(write_dir, suffix="fasta", mode=OVERWRITE)
    dstore.write_citations(data=sample_citations)
    source = pathlib.Path(dstore.source)
    path = shutil.make_archive(
        base_name=str(source),
        format="zip",
        base_dir=source.name,
        root_dir=source.parent,
    )
    zipped = ReadOnlyDataStoreZipped(pathlib.Path(path), suffix="fasta")
    cited = zipped.summary_citations
    assert isinstance(cited, list)
    assert len(cited) == 2


def test_write_citations_zipped_raises(zipped_basic):
    zipped = ReadOnlyDataStoreZipped(zipped_basic, suffix="fasta")
    with pytest.raises(TypeError, match="read only"):
        zipped.write_citations(data=(None,))


def test_old_directory_store_without_citations(fasta_dir):
    """Opening a directory store created before citations were added works."""
    # fasta_dir has .fasta files but no .citations file
    dstore = DataStoreDirectory(fasta_dir, suffix="fasta", mode=READONLY)
    assert dstore._load_citations() == []
    cited = dstore.summary_citations
    assert isinstance(cited, list)
    assert len(cited) == 0


_dict_types = [dict]
if UnionDict is not None:
    _dict_types.append(UnionDict)
_types = tuple(product(_dict_types, (str, Path)))


@pytest.mark.parametrize(("container_type", "source_stype"), _types)
def test_get_data_source_dict(container_type, source_stype):
    """handles case where input is dict (sub)class instance with top level source key"""
    value = source_stype("some/path.txt")
    data = container_type(source=value)
    got = get_data_source(data)
    assert got == "path.txt"


@pytest.mark.parametrize("klass", [str, Path])
def test_get_data_source_seqcoll(klass):
    """handles case where input is sequence collection object"""
    from cogent3 import make_unaligned_seqs

    value = klass("some/path.txt")
    obj = make_unaligned_seqs(
        {"seq1": "ACGG"},
        moltype="dna",
        info={"random_key": 1234},
        source=value,
    )
    got = get_data_source(obj)
    assert got == "path.txt"



import contextlib
import inspect
import json
import re
import sys
import textwrap
import time
import traceback
import types
import typing
from collections.abc import Callable, Iterable, Iterator
from copy import copy, deepcopy
from enum import Enum
from pathlib import Path
from typing import Any, ClassVar, Generic, Literal, Self, TypeVar, overload
from uuid import uuid4

from citeable import Citation
from scitrack import CachingLogger  # type: ignore[import-untyped]
from typeguard import TypeCheckError, check_type

from scinexus import typing as snx_typing
from scinexus._version import __version__
from scinexus.deserialise import register_deserialiser
from scinexus.misc import docstring_to_summary_rest, get_object_provenance
from scinexus.progress import Progress, get_progress
from scinexus.typing import (
    check_type_compatibility,
    get_type_display_names,
    resolve_type_hint,
)
from scinexus.warning import deprecated_callable

from .data_store import (
    DataMember,
    DataStoreABC,
    get_id_from_source,
)

_builtin_seqs = list, set, tuple

T = TypeVar("T")
R = TypeVar("R")
_T = TypeVar("_T")


def _make_logfile_name(process: object) -> str:
    text = re.split(r"\s+\+\s+", str(process))
    parts = []
    for part in text:
        if part.find("(") >= 0:
            part = part[: part.find("(")]
        parts.append(part)
    result = "-".join(parts)
    uid = str(uuid4())
    return f"{result}-{uid[:8]}.log"


def _get_origin(origin: typing.Any) -> str:
    return origin if isinstance(origin, str) else origin.__class__.__name__


class NotCompletedType(Enum):
    ERROR = "ERROR"
    FAIL = "FAIL"
    BUG = "BUG"


class NotCompleted(int):
    """results that failed to complete"""

    type: NotCompletedType
    origin: str
    message: str
    source: str | None
    _persistent: tuple[tuple[Any, ...], dict[str, Any]]

    def __new__(
        cls,
        type_: NotCompletedType | str,
        origin: typing.Any,
        message: str,
        source: typing.Any = None,
    ) -> Self:
        """
        Parameters
        ----------
        type_
            the category of failure, e.g. NotCompletedType.ERROR
        origin
            where the instance was created, can be an instance
        message
            descriptive message, succinct traceback
        source
            the data operated on that led to this result. May be a string
            or an instance with ``.source`` or ``.info.source`` attributes.
        """
        type_ = NotCompletedType(type_)
        origin = _get_origin(origin)
        try:
            source = get_id_from_source()(source)
        except Exception:  # noqa: BLE001
            source = None
        result = int.__new__(cls, False)  # noqa: FBT003
        result._persistent = (type_.value, origin, message), {"source": source}  # noqa: SLF001
        result.type = type_
        result.origin = origin
        result.message = message
        result.source = source
        return result

    def __getnewargs_ex__(self) -> tuple[tuple[Any, ...], dict[str, Any]]:
        return self._persistent[0], self._persistent[1]

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        name = self.__class__.__name__
        source = self.source or "Unknown"
        return f'{name}(type={self.type.value}, origin={self.origin}, source="{source}", message="{self.message}")'

    def to_rich_dict(self) -> dict[str, Any]:
        """returns components for to_json"""
        return {
            "type": get_object_provenance(self),
            "not_completed_construction": {
                "args": self._persistent[0],
                "kwargs": self._persistent[1],
            },
            "version": __version__,
        }

    def to_json(self) -> str:
        """returns json string"""
        return json.dumps(self.to_rich_dict())


class AppType(Enum):
    LOADER = "loader"
    WRITER = "writer"
    GENERIC = "generic"
    NON_COMPOSABLE = "non_composable"


# Aliases to use Enum easily
LOADER = AppType.LOADER
WRITER = AppType.WRITER
GENERIC = AppType.GENERIC
NON_COMPOSABLE = AppType.NON_COMPOSABLE


def _get_raw_hints(
    main_func: Callable[..., Any], min_params: int
) -> tuple[object, object]:
    _no_value = inspect.Parameter.empty
    params = inspect.signature(main_func)
    if len(params.parameters) < min_params:
        msg = f"{main_func.__name__!r} must have at least {min_params} input parameters"
        raise ValueError(
            msg,
        )
    # annotation for first parameter other than self, params.parameters is an orderedDict
    first_param_type = [p.annotation for p in params.parameters.values()][
        min_params - 1
    ]
    return_type = params.return_annotation
    if return_type is _no_value:
        msg = "must specify type hint for return type"
        raise TypeError(msg)
    if first_param_type is _no_value:
        msg = "must specify type hint for first parameter"
        raise TypeError(msg)

    if first_param_type is None:
        msg = "NoneType invalid type for first parameter"
        raise TypeError(msg)
    if return_type is None:
        msg = "NoneType invalid type for return value"
        raise TypeError(msg)

    if isinstance(first_param_type, str):
        msg = (
            "Apps do not yet support string type hints "
            "(such as those caused by __future__ annotations). "
            f"Bad type hint: {first_param_type}"
        )
        raise NotImplementedError(msg)
    if isinstance(return_type, str):
        msg = (
            "Apps do not yet support string type hints "
            "(such as those caused by __future__ annotations). "
            f"Bad type hint: {return_type}"
        )
        raise NotImplementedError(msg)

    return first_param_type, return_type


def _get_main_hints(klass: type[Any]) -> tuple[object, object]:
    """return raw type hints for main method

    Returns
    -------
    (first_param_type_hint, return_type_hint)
    """
    # Check klass.main exists and is type method
    main_func = getattr(klass, "main", None)
    if (
        main_func is None
        or not inspect.isclass(klass)
        or not inspect.isfunction(main_func)
    ):
        msg = f"must define a callable main() method in {klass.__name__!r}"
        raise ValueError(msg)

    first_param_type, return_type = _get_raw_hints(main_func, 2)
    return first_param_type, return_type


def _set_hints(
    main_meth: Callable[..., Any], first_param_type: object, return_type: object
) -> Callable[..., Any]:
    """adds type hints to main"""
    main_meth.__annotations__["arg"] = first_param_type
    main_meth.__annotations__["return"] = return_type
    return main_meth


class source_proxy(Generic[_T]):
    """wraps an object to track its source through app pipelines"""

    __slots__ = ("_obj", "_src", "_uuid")

    def __init__(self, obj: _T) -> None:
        self._obj = obj
        self._src = obj
        self._uuid = uuid4()

    def __hash__(self) -> int:
        return hash(self._uuid)

    @property
    def obj(self) -> _T:
        return self._obj

    def set_obj(self, obj: typing.Any) -> None:
        self._obj = obj

    @property
    def source(self) -> typing.Any:
        """origin of this object"""
        return self._src

    @source.setter
    def source(self, src: typing.Any) -> None:
        # need to check whether src is hashable, how to cope if it isn't?
        # might need to make this instance hashable perhaps using a uuid?
        self._src = src

    @property
    def uuid(self) -> str:
        """unique identifier for this object"""
        return str(self._uuid)

    def __getattr__(self, name: str) -> typing.Any:
        return getattr(self._obj, name)

    def __setattr__(self, name: str, value: typing.Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            setattr(self._obj, name, value)

    def __bool__(self) -> bool:
        return bool(self._obj)

    def __repr__(self) -> str:
        return self.obj.__repr__()

    def __str__(self) -> str:
        return self.obj.__str__()

    def __eq__(self, other: object) -> bool:
        return self.obj.__eq__(other)

    def __len__(self) -> int:
        return self.obj.__len__()  # type: ignore[attr-defined]

    # pickling induces infinite recursion on python 3.10
    # only on Windows, so implementing the following methods explicitly
    def __getstate__(self) -> tuple[Any, Any, Any]:
        return self._obj, self._src, self._uuid

    def __setstate__(self, state: tuple[Any, Any, Any]) -> None:
        self._obj, self._src, self._uuid = state


def _proxy_input(dstore: Iterable[Any]) -> list[source_proxy[Any]]:
    inputs = []
    for e in dstore:
        if not e:
            continue
        if not isinstance(e, source_proxy):
            e = e if hasattr(e, "source") else source_proxy(e)
        inputs.append(e)

    return inputs


GetIdFuncType = typing.Callable[[source_proxy[Any] | snx_typing.HasSource], str | None]


class propagate_source:
    """retains result association with source

    Notes
    -----
    Returns the unwrapped result if it has a .source attribute,
    otherwise returns the original source_proxy with the .obj
    updated with result.
    """

    def __init__(self, app: "AppBase[Any, Any]", id_from_source: GetIdFuncType) -> None:
        self.app = app
        self.id_from_source = id_from_source

    def __call__(
        self, value: source_proxy[Any] | snx_typing.HasSource
    ) -> snx_typing.HasSource:
        if not isinstance(value, source_proxy):
            return self.app(value)

        result = self.app(value.obj)
        if self.id_from_source(result):
            return result

        value.set_obj(result)
        return value


# Forbidden methods per app kind
_FORBIDDEN_BASE = frozenset(
    {
        "__call__",
        "__repr__",
        "__str__",
        "__new__",
        "__copy__",
        "__eq__",
        "_validate_data_type",
        "as_completed",
        "check_data_type",
        "_get_citations",
        "citations",
        "bib",
    }
)
_FORBIDDEN_COMPOSABLE = _FORBIDDEN_BASE | frozenset(
    {
        "__add__",
        "disconnect",
        "input",
    }
)
_FORBIDDEN_WRITER = _FORBIDDEN_COMPOSABLE | frozenset(
    {
        "apply_to",
        "set_logger",
    }
)


def _init_subclass_setup(
    cls: Any,
    app_type: AppType | Literal["loader", "writer", "generic", "non_composable"],
    skip_not_completed: bool,
    cite: Citation | None,
) -> None:
    """Shared setup logic for __init_subclass__ and define_app."""
    app_type = AppType(app_type)

    if "__slots__" in cls.__dict__:
        msg = "slots are not currently supported"
        raise NotImplementedError(msg)

    if app_type is WRITER:
        forbidden = _FORBIDDEN_WRITER
    elif app_type is not NON_COMPOSABLE:
        forbidden = _FORBIDDEN_COMPOSABLE
    else:
        forbidden = _FORBIDDEN_BASE

    if (
        app_type is not NON_COMPOSABLE
        and "input" in cls.__dict__
        and cls.__dict__["input"] is not None
    ):
        msg = f"remove 'input' attribute in {cls.__name__!r}, reserved by the app framework"
        raise TypeError(msg)

    for meth in forbidden:
        if meth in cls.__dict__:
            val = cls.__dict__[meth]
            if isinstance(val, staticmethod):
                val = val.__func__
            if inspect.isfunction(val) or isinstance(val, property):
                msg = f"remove {meth!r} in {cls.__name__!r}, reserved by the app framework"
                raise TypeError(msg)

    raw_input, raw_return = _get_main_hints(cls)
    mod = sys.modules.get(cls.__module__) if cls.__module__ else None
    module_globals = vars(mod) if mod else {}
    cls._input_type = resolve_type_hint(raw_input, module_globals)
    cls._return_type = resolve_type_hint(raw_return, module_globals)
    cls.app_type = app_type
    cls._skip_not_completed = skip_not_completed
    cls._check_data_type = True
    cls._cite = cite
    cls._source_wrapped = None

    if app_type is not LOADER:
        cls.input = None


class AppBase(Generic[T, R]):
    """Base for all app types. Provides __call__, __repr__, etc.

    Raises
    ------
    TypeError
        If a subclass defines any method name reserved by the app framework.
        The reserved names depend on the app type.

        All app types: ``__call__``, ``__repr__``, ``__str__``, ``__new__``,
        ``__copy__``, ``__eq__``, ``_validate_data_type``, ``as_completed``,
        ``check_data_type``, ``_get_citations``, ``citations``, ``bib``.

        Composable apps (GENERIC, LOADER, WRITER) additionally:
        ``__add__``, ``disconnect``, ``input``.

        Writer apps additionally: ``apply_to``, ``set_logger``.
    """

    _is_intermediate_base: bool = False
    _skip_not_completed: bool
    _check_data_type: bool
    _source_wrapped: propagate_source | None
    _cite: Citation | None
    _input_type: type
    _return_type: type
    _init_vals: dict[str, Any]
    app_type: AppType
    input: Any
    main: Callable[..., Any]

    def __init_subclass__(
        cls,
        skip_not_completed: bool = True,
        cite: Citation | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(**kwargs)
        # Skip setup for intermediate bases and classes built by define_app
        if "_is_intermediate_base" in cls.__dict__ or getattr(
            cls, "_define_app_pending", False
        ):
            return
        app_type = getattr(cls, "_default_app_type", GENERIC)
        _init_subclass_setup(cls, app_type, skip_not_completed, cite)

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        obj = object.__new__(cls)

        if hasattr(cls, "_func_sig"):
            # we have a decorated function, the first parameter in the signature
            # is not given to constructor, so we create a new signature excluding that one
            params = cls._func_sig.parameters  # type: ignore[attr-defined]
            init_sig = inspect.Signature(parameters=list(params.values())[1:])
            bargs = init_sig.bind_partial(*args, **kwargs)
        else:
            init_sig = inspect.signature(cls.__init__)
            bargs = init_sig.bind_partial(cls, *args, **kwargs)
        bargs.apply_defaults()
        init_vals = bargs.arguments
        init_vals.pop("self", None)

        obj._init_vals = init_vals
        return obj

    def __copy__(self) -> Self:
        new = object.__new__(type(self))
        new.__dict__.update(self.__dict__)
        return new

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return all(
            v is other.__dict__.get(k) for k, v in self.__dict__.items() if k != "input"
        )

    __hash__ = None  # type: ignore[assignment]

    @property
    def check_data_type(self) -> bool:
        """toggle whether the type of input data matches the defined compatible types

        Notes
        -----
        If False, and an invalid data type is passed, the error will
        still be caught, but in a potentially less informative way. For instance
        "'NoneType' object has no attribute 'blah'"
        """
        return self._check_data_type

    @check_data_type.setter
    def check_data_type(self, value: bool) -> None:
        self._check_data_type = value
        head = getattr(self, "input", None)
        while head is not None:
            head._check_data_type = value
            head = getattr(head, "input", None)

    def __call__(self, val: T, *args: Any, **kwargs: Any) -> R | NotCompleted:
        if val is None:
            return NotCompleted(
                NotCompletedType.ERROR, self, "unexpected input value None", source=val
            )

        if isinstance(val, NotCompleted) and self._skip_not_completed:
            return val

        if self.app_type is not LOADER and self.input:  # passing to connected app
            val = self.input(val, *args, **kwargs)
            if isinstance(val, NotCompleted) and self._skip_not_completed:
                return val

        if self._check_data_type:
            type_checked = self._validate_data_type(val)
            if not type_checked:
                return type_checked  # type: ignore[return-value]

        try:
            result = self.main(val, *args, **kwargs)
        except Exception:
            result = NotCompleted(
                NotCompletedType.ERROR, self, traceback.format_exc(), source=val
            )

        if result is None:
            result = NotCompleted(
                NotCompletedType.BUG, self, "unexpected output value None", source=val
            )
        return result

    def __repr__(self) -> str:
        val = f"{self.input!r} + " if self.app_type is not LOADER and self.input else ""
        all_args = {**self._init_vals}
        args_items = all_args.pop("args", None)
        data = ", ".join(f"{v!r}" for v in args_items) if args_items else ""
        kwargs_items = all_args.pop("kwargs", None)
        data += (
            ", ".join(f"{k}={v!r}" for k, v in kwargs_items.items())
            if kwargs_items
            else ""
        )
        data += ", ".join(f"{k}={v!r}" for k, v in all_args.items())
        data = f"{val}{self.__class__.__name__}({data})"
        return textwrap.fill(
            data, width=80, break_long_words=False, break_on_hyphens=False
        )

    __str__ = __repr__

    def _validate_data_type(self, data: Any) -> bool | NotCompleted:
        """checks data type matches defined compatible types using typeguard"""
        if isinstance(data, NotCompleted):
            if self._skip_not_completed:
                return data
            # skip_not_completed=False means the app handles NotCompleted itself
            return True

        if isinstance(data, source_proxy):
            data = data.obj

        if isinstance(data, _builtin_seqs) and len(data) == 0:
            return NotCompleted(
                NotCompletedType.ERROR, self, message="empty data", source=data
            )

        try:
            check_type(data, self._input_type)
            return True
        except TypeCheckError:
            class_name = data.__class__.__name__
            expected = get_type_display_names(self._input_type)
            msg = f"invalid data type, '{class_name}' not in {', '.join(sorted(expected))}"
            return NotCompleted(NotCompletedType.ERROR, self, message=msg, source=data)

    def as_completed(
        self,
        dstore: DataStoreABC | Iterable[Any] | str,
        parallel: bool = False,
        par_kw: dict[str, Any] | None = None,
        id_from_source: GetIdFuncType | None = None,
        show_progress: bool | Progress = False,
    ) -> Iterator[Any]:
        """invokes self composable function on the provided data store

        Parameters
        ----------
        dstore
            a path, list of paths, or DataStore to which the process will be
            applied.
        parallel
            run in parallel, according to arguments in par_kwargs. If True,
            the last step of the composable function serves as the master
            process, with earlier steps being executed in parallel for each
            member of dstore.
        par_kw
            dict of values for configuring parallel execution.
        id_from_source
            extracts a unique identifier from each input. If not provided,
            defaults to the function registered via
            ``scinexus.data_store.set_id_from_source``, falling back to
            ``scinexus.data_store.get_unique_id``.
        show_progress
            controls progress bar display. Pass ``True`` for the default
            progress bar, ``False`` to disable, or a ``Progress`` instance
            for a custom backend.

        Notes
        -----
        If run in parallel, this instance serves as the master object and
        aggregates results. If run in serial, results are returned in the
        same order as provided.
        """
        if id_from_source is None:
            id_from_source = get_id_from_source()
        if self._source_wrapped is None:
            app = propagate_source(
                self.input if self.app_type is WRITER else self, id_from_source
            )
        else:
            app = (
                self.input._source_wrapped
                if self.app_type is WRITER
                else self._source_wrapped
            )

        if isinstance(dstore, str):
            dstore = [dstore]
        elif isinstance(dstore, DataStoreABC):
            dstore = dstore.completed
        mapped = _proxy_input(dstore)
        if not mapped:
            return (_ for _ in ())

        if parallel:
            from scinexus import parallel as snxpar

            par_kw = par_kw or {}
            to_do: typing.Iterable = snxpar.as_completed(app, mapped, **par_kw)
        else:
            to_do = map(app, mapped)

        progress = get_progress(show_progress)
        return progress(to_do, total=len(mapped))

    def _get_citations(self) -> tuple[Citation, ...]:
        """Return citations for this app and all composed input apps."""
        seen: set[Citation] = set()
        result: list[Citation] = []

        if self._cite is not None:
            self._cite.app = self.__class__.__name__
            seen.add(self._cite)
            result.append(self._cite)

        head = getattr(self, "input", None)
        while head is not None:
            if head._cite is not None and head._cite not in seen:
                head._cite.app = head.__class__.__name__
                seen.add(head._cite)
                result.append(head._cite)
            head = getattr(head, "input", None)

        return tuple(result)

    @property
    def citations(self) -> tuple[Citation, ...]:
        """Citations for this app and all composed input apps."""
        return self._get_citations()

    @property
    def bib(self) -> str:
        """BibTeX formatted string of citations for this app and all composed input apps."""
        return "\n\n".join(str(cite) for cite in self.citations)


class ComposableApp(AppBase[T, R]):
    """Adds __add__ for LOADER/GENERIC.

    Raises
    ------
    TypeError
        If a subclass defines any method name reserved by the app framework:
        ``__call__``, ``__repr__``, ``__str__``, ``__new__``, ``__copy__``,
        ``__eq__``, ``_validate_data_type``, ``as_completed``,
        ``check_data_type``, ``_get_citations``, ``citations``, ``bib``,
        ``__add__``, ``disconnect``, ``input``.
    """

    _is_intermediate_base: bool = True

    def __add__(self, other: "ComposableApp[Any, Any]") -> "ComposableApp[Any, Any]":
        if getattr(other, "app_type", None) not in {WRITER, LOADER, GENERIC}:
            msg = f"{other!r} is not composable"
            raise TypeError(msg)

        if other == self:
            msg = "cannot add an app to itself"
            raise ValueError(msg)

        # Check order
        if self.app_type is WRITER:
            msg = "Left hand side of add operator must not be of type writer"
            raise TypeError(msg)
        if other.app_type is LOADER:
            msg = "Right hand side of add operator must not be of type loader"
            raise TypeError(msg)

        if not check_type_compatibility(self._return_type, other._input_type):
            self_names = get_type_display_names(self._return_type)
            other_names = get_type_display_names(other._input_type)
            msg = (
                f"{self.__class__.__name__!r} return_type {self_names} "
                f"incompatible with {other.__class__.__name__!r} input "
                f"type {other_names}"
            )
            raise TypeError(msg)
        result = copy(other)
        result.input = copy(self)
        return result

    @deprecated_callable(
        version="2026.9",
        reason="no longer required",
        is_discontinued=True,
    )
    def disconnect(self) -> None:  # pragma: no cover
        """Deprecated. No longer required since composition uses copies."""


class WriterApp(ComposableApp[T, R]):
    """Adds apply_to and set_logger for WRITER.

    Raises
    ------
    TypeError
        If a subclass defines any method name reserved by the app framework:
        ``__call__``, ``__repr__``, ``__str__``, ``__new__``, ``__copy__``,
        ``__eq__``, ``_validate_data_type``, ``as_completed``,
        ``check_data_type``, ``_get_citations``, ``citations``, ``bib``,
        ``__add__``, ``disconnect``, ``input``, ``apply_to``, ``set_logger``.
    """

    _is_intermediate_base: bool = True
    _default_app_type: ClassVar[AppType] = WRITER
    data_store: DataStoreABC
    logger: CachingLogger | None

    def apply_to(
        self,
        dstore: DataStoreABC | Iterable[Any] | str | Path,
        id_from_source: GetIdFuncType | None = None,
        parallel: bool = False,
        par_kw: dict[str, Any] | None = None,
        logger: bool | CachingLogger = True,
        cleanup: bool = True,
        show_progress: bool | Progress = False,
    ) -> DataStoreABC:
        """invokes self composable function on the provided data store

        Parameters
        ----------
        dstore
            a path, list of paths, or DataStore to which the process will be
            applied.
        id_from_source
            makes the unique identifier from elements of dstore that will be
            used for writing results. If not provided, defaults to the
            function registered via
            ``scinexus.data_store.set_id_from_source``, falling back to
            ``scinexus.data_store.get_unique_id``.
        parallel
            run in parallel, according to arguments in par_kwargs. If True,
            the last step of the composable function serves as the master
            process, with earlier steps being executed in parallel for each
            member of dstore.
        par_kw
            dict of values for configuring parallel execution.
        logger
            Controls logging. If True (default), a CachingLogger is created
            automatically. If False, logging is disabled. A CachingLogger
            instance can be passed directly.
        cleanup
            after copying of log files into the data store, it is deleted
            from the original location
        show_progress
            controls progress bar display

        Returns
        -------
        The output data store instance

        Notes
        -----
        This is an append only function, meaning that if a member already exists
        in self.data_store for an input, it is skipped.

        If run in parallel, this instance spawns workers and aggregates results.
        """
        if id_from_source is None:
            id_from_source = get_id_from_source()
        if self.app_type is WRITER:
            if self.input is None:
                msg = "writer app has no composed input"
                raise RuntimeError(msg)
            self.input._source_wrapped = propagate_source(self.input, id_from_source)
            self._source_wrapped = propagate_source(self, id_from_source)

        if isinstance(dstore, str | Path):  # one filename
            dstore = [dstore]
        elif isinstance(dstore, DataStoreABC):
            dstore = dstore.completed

        # TODO this should fail if somebody provides data that cannot produce a unique_id
        inputs = {}
        for m in dstore:
            input_id = Path(m.unique_id) if isinstance(m, DataMember) else m
            input_id = id_from_source(input_id)  # type: ignore[arg-type]
            if input_id in inputs or not input_id:
                msg = f"non-unique identifier {input_id!r} detected in data"
                raise ValueError(msg)
            if input_id in self.data_store:
                # we are assuming that this query returns True only when
                # an input_id is completed, we will not hit this if not_completed
                continue
            inputs[input_id] = m

        if (
            not dstore
        ):  # this should just return datastore, because if all jobs are done!
            msg = "dstore is empty"
            raise ValueError(msg)

        self.set_logger(logger)
        active_logger: CachingLogger | None = self.logger
        if active_logger:
            start = time.time()
            active_logger.log_message(str(self), label="composable function")
            active_logger.log_versions(["scinexus"])

        proxied = _proxy_input(inputs.values())
        for result in self.as_completed(
            proxied,
            parallel=parallel,
            par_kw=par_kw,
            show_progress=show_progress,
        ):
            member = self.main(
                data=getattr(result, "obj", result),
                identifier=id_from_source(result),  # type: ignore[arg-type]
            )
            if active_logger:
                md5 = getattr(member, "md5", None)
                active_logger.log_message(str(member), label="output")
                if md5:
                    active_logger.log_message(md5, label="output md5sum")

        if active_logger:
            taken = time.time() - start
            active_logger.log_message(f"{taken}", label="TIME TAKEN")
            log_file_path = Path(active_logger.log_file_path)
            active_logger.shutdown()
            self.data_store.write_log(
                unique_id=log_file_path.name,
                data=log_file_path.read_text(),
            )
            if cleanup:
                log_file_path.unlink(missing_ok=True)

        # write citations
        self.data_store.write_citations(data=self.citations)

        return self.data_store

    def set_logger(self, logger: bool | CachingLogger = True) -> None:
        if logger is False:
            self.logger = None
            return
        if logger is True:
            logger = CachingLogger(create_dir=True)
        if not isinstance(logger, CachingLogger):
            msg = f"logger must be of type CachingLogger not {type(logger)}"
            raise TypeError(msg)
        if not logger.log_file_path:
            src = Path(self.data_store.source).parent  # type: ignore[attr-defined]
            logger.log_file_path = str(src / _make_logfile_name(self))
        self.logger = logger


class LoaderApp(ComposableApp[T, R]):
    """Intermediate base class for LOADER apps.

    Subclasses of ``LoaderApp`` are automatically assigned
    ``app_type=LOADER``. Loaders sit at the start of a composed pipeline
    and have no ``input`` attribute.

    Raises
    ------
    TypeError
        If a subclass defines any method name reserved by the app framework:
        ``__call__``, ``__repr__``, ``__str__``, ``__new__``, ``__copy__``,
        ``__eq__``, ``_validate_data_type``, ``as_completed``,
        ``check_data_type``, ``_get_citations``, ``citations``, ``bib``,
        ``__add__``, ``disconnect``, ``input``.

    Examples
    --------
    Define a loader by inheritance::

        class my_loader(LoaderApp):
            def main(self, path):
                return path
    """

    _is_intermediate_base: bool = True
    _default_app_type: ClassVar[AppType] = LOADER


class NonComposableApp(AppBase[T, R]):
    """Intermediate base class for NON_COMPOSABLE apps.

    Subclasses of ``NonComposableApp`` are automatically assigned
    ``app_type=NON_COMPOSABLE``. Non-composable apps cannot participate
    in pipeline composition via ``+``.

    Raises
    ------
    TypeError
        If a subclass defines any method name reserved by the app framework:
        ``__call__``, ``__repr__``, ``__str__``, ``__new__``, ``__copy__``,
        ``__eq__``, ``_validate_data_type``, ``as_completed``,
        ``check_data_type``, ``_get_citations``, ``citations``, ``bib``.

    Examples
    --------
    Define a non-composable app by inheritance::

        class my_app(NonComposableApp):
            def main(self, val):
                return val
    """

    _is_intermediate_base: bool = True
    _default_app_type: ClassVar[AppType] = NON_COMPOSABLE


def _class_from_func(func: Callable[..., Any]) -> type[Any]:
    """make a class based on func

    Notes
    -----
    produces a class consistent with the necessary properties for
    the define_app class decorator.

    func becomes a static method on the class
    """

    # these methods MUST be in function scope so that separate instances are
    # created for each decorated function
    def _init(self: Any, *args: Any, **kwargs: Any) -> None:
        self._args = args
        self._kwargs = kwargs
        self._source_wrapped = None

    def _main(self: Any, arg: Any, *args: Any, **kwargs: Any) -> Any:
        kw_args = deepcopy(self._kwargs)
        kw_args = {**kw_args, **kwargs}
        args = (arg, *args, *deepcopy(self._args))
        bound = self._func_sig.bind(*args, **kw_args)
        return self._user_func(**bound.arguments)

    module = func.__module__  # to be assigned to the generated class
    sig = inspect.signature(func)
    class_name = func.__name__
    _main = _set_hints(_main, *_get_raw_hints(func, 1))
    summary, body = docstring_to_summary_rest(func.__doc__ or "")
    func.__doc__ = None

    _class_dict = {"__init__": _init, "main": _main, "_user_func": staticmethod(func)}

    for method_name, method in _class_dict.items():
        method.__name__ = method_name  # type: ignore[attr-defined]
        method.__qualname__ = f"{class_name}.{method_name}"  # type: ignore[attr-defined]

    result = types.new_class(class_name, (), exec_body=lambda x: x.update(_class_dict))
    result.__module__ = module  # necessary for pickle support
    result._func_sig = sig  # type: ignore[attr-defined]
    result.__doc__ = summary
    result.__init__.__doc__ = body  # type: ignore[misc]
    return result


def _fix_super_class_cells(
    old_klass: type[Any],
    new_klass: type[Any],
) -> None:
    """Update __class__ closure cells so zero-arg super() works in the new class.

    When define_app rebuilds a class via types.new_class, methods copied from
    the original class retain closure cells pointing to the old class. Python's
    zero-arg super() uses the __class__ cell, so it fails when the instance is
    of the new class. This function patches those cells to point to new_klass.
    """
    for attr in new_klass.__dict__.values():
        fn = attr if isinstance(attr, types.FunctionType) else None
        if fn is None:
            # unwrap classmethod / staticmethod
            fn = getattr(attr, "__func__", None)
        if not isinstance(fn, types.FunctionType):
            continue
        closure = fn.__closure__
        if closure is None:
            continue
        freevars = fn.__code__.co_freevars
        for i, name in enumerate(freevars):
            if name == "__class__":
                with contextlib.suppress(ValueError):
                    if closure[i].cell_contents is old_klass:
                        closure[i].cell_contents = new_klass


@overload
def define_app(
    klass: type[Any] | Callable[..., Any],
) -> type[ComposableApp[Any, Any]]: ...


@overload
def define_app(  # type: ignore[overload-overlap]
    klass: None = None,
    *,
    app_type: Literal[AppType.NON_COMPOSABLE, "non_composable"],
    skip_not_completed: bool = ...,
    cite: Citation | None = ...,
) -> Callable[[type[Any] | Callable[..., Any]], type[AppBase[Any, Any]]]: ...


@overload
def define_app(  # type: ignore[overload-overlap]
    klass: None = None,
    *,
    app_type: Literal[AppType.WRITER, "writer"],
    skip_not_completed: bool = ...,
    cite: Citation | None = ...,
) -> Callable[[type[Any] | Callable[..., Any]], type[WriterApp[Any, Any]]]: ...


@overload
def define_app(
    klass: None = None,
    *,
    app_type: AppType | Literal["loader", "writer", "generic", "non_composable"] = ...,
    skip_not_completed: bool = ...,
    cite: Citation | None = ...,
) -> Callable[[type[Any] | Callable[..., Any]], type[ComposableApp[Any, Any]]]: ...


def define_app(
    klass: type[Any] | Callable[..., Any] | None = None,
    *,
    app_type: AppType
    | Literal["loader", "writer", "generic", "non_composable"] = GENERIC,
    skip_not_completed: bool = True,
    cite: Citation | None = None,
) -> type[Any]:
    """decorator for building callable apps

    Parameters
    ----------
    klass
        either a class or a function. If a function, it is converted to a
        class with the function bound as a static method.
    app_type
        what type of app, typically you just want GENERIC.
    skip_not_completed
        if True (default), NotCompleted instances are returned without being
        passed to the app.
    cite
        a Citation instance describing the software or algorithm. If provided,
        its ``.app`` attribute is set to the class name.

    Raises
    ------
    TypeError
        If the decorated class defines any method name reserved by the app
        framework. The reserved names depend on the app type.

        All app types: ``__call__``, ``__repr__``, ``__str__``, ``__new__``,
        ``__copy__``, ``__eq__``, ``_validate_data_type``, ``as_completed``,
        ``check_data_type``, ``_get_citations``, ``citations``, ``bib``.

        Composable apps (GENERIC, LOADER, WRITER) additionally:
        ``__add__``, ``disconnect``, ``input``.

        Writer apps additionally: ``apply_to``, ``set_logger``.

    Examples
    --------

    An example app definition.

    >>> from scinexus.composable import define_app

    >>> @define_app
    ... class noop:
    ...     def main(self, data: int) -> int:
    ...         return data

    Notes
    -----

    Instances of scinexus apps are callable. If an exception occurs,
    the app returns a ``NotCompleted`` instance with logging information.
    Apps defined with app_type ``LOADER``, ``GENERIC`` or ``WRITER`` can be
    "composed" (summed together) to produce a single callable that
    sequentially invokes the composed apps. For example, the independent
    usage of app instances ``app1`` and ``app2`` as

    ```python { notest }
    app2(app1(data))
    ```

    is equivalent to

    ```python { notest }
    combined = app1 + app2
    combined(data)
    ```

    The ``app_type`` attribute is used to constrain how apps can be composed.
    ``LOADER`` and ``WRITER`` are special cases. If included, a ``LOADER``
    must always be first, e.g.

    ```python { notest }
    app = a_loader + a_generic
    ```

    If included, a ``WRITER`` must always be last, e.g.

    ```python { notest }
    app = a_generic + a_writer
    ```

    Changing the order for either of the above will result in a ``TypeError``.

    There are no constraints on ordering of ``GENERIC`` aside from compatability of
    their input and return types (see below).

    In order to be decorated with ``@define_app`` a class **must**

    - implement a method called ``main``
    - type hint the first argument of ``main``
    - type hint the return type for ``main``

    While you can have more than one argument in ``main``, this is currently not
    supported in composable apps.

    Overlap between the return type hint and first argument hint is required
    for two apps to be composed together.

    ``define_app`` adds a ``__call__`` method which checks an input value prior
    to passing it to ``app.main()`` as a positional argument. The data checking
    results in ``NotCompleted`` being returned immediately, unless
    ``skip_not_completed==False``. If the input value type is consistent with
    the type hint on the first argument of main it is passed to ``app.main()``.
    If it does not match, a new ``NotCompleted`` instance is returned.
    """

    if hasattr(klass, "app_type"):
        msg = (
            f"The class {klass.__name__!r} is already decorated, avoid using "  # type: ignore[union-attr]
            "inheritance from a decorated class."
        )
        raise TypeError(
            msg,
        )

    app_type = AppType(app_type)

    def wrapped(klass: type[Any] | Callable[..., Any]) -> type[Any]:
        if inspect.isfunction(klass):
            klass = _class_from_func(klass)
        if not inspect.isclass(klass):
            msg = f"{klass} is not a class"
            raise ValueError(msg)

        # Select base class based on app_type
        base: type[AppBase[Any, Any]]
        if app_type is WRITER:
            base = WriterApp
        elif app_type is not NON_COMPOSABLE:
            base = ComposableApp
        else:
            base = AppBase

        if "__slots__" in klass.__dict__:
            msg = "slots are not currently supported"
            raise NotImplementedError(msg)

        # Get type hints before rebuilding the class
        raw_input, raw_return = _get_main_hints(klass)

        # Collect the user's class dict (excluding metaclass artefacts)
        original_dict = {
            k: v
            for k, v in klass.__dict__.items()
            if k not in ("__dict__", "__weakref__")
        }
        # Prevent __init_subclass__ from running setup (we do it below)
        original_dict["_define_app_pending"] = True

        # Preserve user-specified bases (excluding object) so super()
        # resolves through the original MRO
        extra_bases = tuple(b for b in klass.__bases__ if b is not object)

        # Recreate class with the base (types.new_class stores
        # parameterised form in __orig_bases__ for type checkers)
        new_klass = types.new_class(
            klass.__name__,
            (base[raw_input, raw_return], *extra_bases),  # type: ignore[index]
            exec_body=lambda ns: ns.update(original_dict),
        )
        new_klass.__module__ = klass.__module__
        new_klass.__qualname__ = klass.__qualname__
        _fix_super_class_cells(klass, new_klass)
        del new_klass._define_app_pending  # type: ignore[attr-defined]

        # Run setup once with the decorator's arguments
        _init_subclass_setup(new_klass, app_type, skip_not_completed, cite)

        return new_klass

    return wrapped(klass) if klass else wrapped  # type: ignore[return-value]


def is_app_composable(obj: object) -> bool:
    """checks whether obj has been decorated by define_app and it's app_type attribute is not NON_COMPOSABLE"""
    return is_app(obj) and obj.app_type is not NON_COMPOSABLE  # type: ignore[attr-defined]


def is_app(obj: object) -> bool:
    """checks whether obj has been decorated by define_app"""
    return hasattr(obj, "app_type")


@register_deserialiser(get_object_provenance(NotCompleted))
def deserialise_not_completed(data: dict[str, Any]) -> NotCompleted:
    """deserialising NotCompletedResult"""
    data.pop("version", None)
    init = data.pop("not_completed_construction")
    args = init.pop("args")
    kwargs = init.pop("kwargs")
    return NotCompleted(*args, **kwargs)



import os
import pathlib
import shutil
import subprocess
import sys

import nox

# on python >= 3.12 this will improve speed of test coverage a lot
if sys.version_info >= (3, 12):
    os.environ["COVERAGE_CORE"] = "sysmon"

_py_versions = range(11, 15)

nox.options.default_venv_backend = "uv"


@nox.session(python=False)
def fmt(session: nox.Session) -> None:
    session.run("ruff", "check", "--fix-only", ".", external=True)
    session.run("ruff", "format", ".", external=True)


@nox.session(python="3.14")
def cogdocs(session: nox.Session) -> None:
    session.install("-e", ".", "--group", "dev")
    cmnd = 'find docs -name "*.md" | xargs uv run --group dev cog -r -I docs/scripts'
    subprocess.run(cmnd, check=True, shell=True)  # noqa: S602


@nox.session(python=[f"3.{v}" for v in _py_versions])
def type_check(session):
    session.install("-e", ".", "--group", "dev")
    session.run("mypy", "src/scinexus/")


@nox.session(python=[f"3.{v}" for v in _py_versions])
def test_types(session):
    session.install("-e", ".")
    session.run("mypy", "src/scinexus/")


@nox.session(python=[f"3.{v}" for v in _py_versions])
def test(session):
    session.install("-e", ".", "--group", "dev")
    session.run("uv", "pip", "list")
    # doctest modules within scinexus
    session.chdir("src/scinexus")
    session.run(
        "pytest",
        "-s",
        "-x",
        "--doctest-modules",
        ".",
    )

    session.chdir("../../tests")
    session.run(
        "pytest",
        "-s",
        "-x",
        "-m",
        "not slow and not mpi",
        *session.posargs,
    )


@nox.session(python=[f"3.{v}" for v in _py_versions])
def testmpi(session):
    session.install("-e", ".[mpi]", "--group", "dev")
    session.chdir("tests")
    py = pathlib.Path(session.bin_paths[0]) / "python"
    session.run(
        "mpiexec",
        "--oversubscribe",
        "-n",
        "4",
        str(py),
        "-m",
        "mpi4py.futures",
        "-m",
        "pytest",
        "-s",
        "-x",
        "-m",
        "mpi",
        *session.posargs,
        external=True,
    )


@nox.session(python=[f"3.{v}" for v in _py_versions])
def testcov(session):
    session.install("-e", ".", "--group", "dev")
    cover_mpi = shutil.which("mpiexec") is not None
    if cover_mpi:
        session.install("-e", ".[mpi]")

    cov_file = str(pathlib.Path.cwd() / ".coverage")
    session.env["COVERAGE_FILE"] = cov_file
    session.run("coverage", "erase")

    base = ["coverage", "run", "--source=scinexus"]

    # mypy via API wrapper so coverage traces plugin hook execution
    session.run(*base, "scripts/run_mypy_cov.py", "--no-incremental", "src/scinexus/")

    # doctests
    session.chdir("src/scinexus")
    session.run(*base, "--append", "-m", "pytest", "-s", "-x", "--doctest-modules", ".")

    # unit tests
    session.chdir("../../tests")
    session.run(
        *base,
        "--append",
        "-m",
        "pytest",
        "-s",
        "-x",
        "-m",
        "not mpi",
    )

    # MPI tests when mpiexec is available
    if cover_mpi:
        py = pathlib.Path(session.bin_paths[0]) / "python"
        session.run(
            "mpiexec",
            "--oversubscribe",
            "-n",
            "4",
            str(py),
            "-m",
            "mpi4py.futures",
            "-m",
            *base,
            "--append",
            "-m",
            "pytest",
            "-s",
            "-x",
            "-m",
            "mpi",
            external=True,
        )

    session.chdir("..")
    session.run("coverage", "report")
    i = 0
    while i < len(session.posargs):
        fmt = session.posargs[i]
        if fmt == "html":
            session.run("coverage", fmt, external=True)
            i += 1
            continue

        o_name = session.posargs[i + 1]
        session.run("coverage", fmt, o_name, external=True)
        i += 2


@nox.session(python=[f"3.{v}" for v in _py_versions])
def test_docs(session):
    session.install("-e", ".", "--group", "dev")
    session.run("uv", "pip", "list")
    # doctest modules within scinexus
    session.chdir("docs")
    session.run(
        "pytest",
        "--markdown-docs",
        "-m",
        "markdown-docs",
        "-x",
        ".",
        "--ignore",
        "scripts",
        *session.posargs,
    )



"""Selected utility functions."""

from __future__ import annotations

import inspect
import re
from typing import TYPE_CHECKING, ParamSpec, TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable


P = ParamSpec("P")
R = TypeVar("R")


def get_object_provenance(obj: object) -> str:
    """returns string of complete object provenance"""
    # algorithm inspired by Greg Baacon's answer to
    # https://stackoverflow.com/questions/2020014/get-fully-qualified-class
    # -name-of-an-object-in-python
    if isinstance(obj, type) or inspect.isfunction(obj):
        mod = obj.__module__
        name = obj.__name__
    else:
        mod = obj.__class__.__module__
        name = obj.__class__.__name__

    return name if mod is None or mod == "builtins" else f"{mod}.{name}"


def extend_docstring_from(
    source: object, pre: bool = False
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    def docstring_inheriting_decorator(dest: Callable[P, R]) -> Callable[P, R]:
        parts = [source.__doc__ or "", dest.__doc__ or ""]
        # trim leading/trailing blank lines from parts
        for i, part in enumerate(parts):
            lines = part.split("\n")
            if not lines[0].strip():
                lines.pop(0)
            if lines and not lines[-1].strip():
                lines.pop(-1)

            parts[i] = "\n".join(lines)

        if pre:
            parts.reverse()
        dest.__doc__ = "\n".join(parts)
        return dest

    return docstring_inheriting_decorator


_doc_block = re.compile(
    r"^\s*(Parameters|Notes|Raises)",
    flags=re.IGNORECASE | re.MULTILINE,
)


def docstring_to_summary_rest(text: str) -> tuple[str, str]:
    """separates the summary at the start of a docstring from the rest

    Notes
    -----
    Assumes numpydoc style.
    """
    if not text:
        return "", ""

    pos = _doc_block.search(text)
    if pos is None:
        return text, ""

    summary = text[: pos.start()].rstrip()
    text = text[pos.start() :]
    return summary, text.lstrip("\n").rstrip(" ")


def in_jupyter() -> bool:
    """whether code is being executed within a jupyter notebook"""
    val = True
    try:
        # primitive approach, just check whether the following function
        # is in the namespace
        get_ipython  # type: ignore[name-defined]  # noqa: B018
    except NameError:
        val = False

    return val
